# -*- coding: utf-8 -*-

import json
import multiprocessing
import pathlib
import queue
import random
import threading
import time
from io import BytesIO
from string import Template

import numpy as np
import requests
from PIL import Image
from PySide2.QtGui import QPixmap

from Osmapy.TileLoader.Tile import Tile
from Osmapy.utils.config import config


class TileLoader:
    """ Class to load slippy tiles in a LIFO queue with workers. The tiles are cached and a cache database with a dict
    is maintained. The Tile Usage Policy of OSM is followed https://operations.osmfoundation.org/policies/tiles/.
    """

    def __init__(self, viewer, config_id):
        """ The constructor needs a viewer reference to update the map as soon as the slippy tile is loaded.

        Args:
            viewer (Viewer): viewer object where the slippy tile should be shown
        """
        self.name = config.slippy_tiles[config_id].name
        self.urls = config.slippy_tiles[config_id].urls

        self.path_cache = pathlib.Path(__file__).parent / pathlib.Path(f"../../cache/{self.name}")
        self.viewer = viewer

        self.cache_json = self.load_cache_json()

        self.queue = queue.LifoQueue()
        self.lock = multiprocessing.Lock()
        for _ in range(min(2, multiprocessing.cpu_count())):    # only two download threads are allowed
            threading.Thread(target=self.worker, daemon=True).start()

    def worker(self):
        """ Worker which downloads the tile, updates the cache database and saves the image. After this processed is
        finished the viewer which requested the image is updated.
        """
        while True:
            tile = self.queue.get()
            # If an error occurse during the loading process the worker should't be blocked. The loading process is
            # tried again later, because the status of the tile in the cache database is still 'loading'.
            try:
                osm_tile_url = random.choice(self.urls)  # randomly chose one of the servers in the list
                request = Template(osm_tile_url)
                request = request.substitute(zoom=tile.zoom, int_xtile=tile.int_xtile, int_ytile=tile.int_ytile)
                # According to the OSM Tile Usage Policy an User-Agent is set
                headers = {"User-Agent": config.user_agent}
                response = requests.get(request, headers=headers)
                image = Image.open(BytesIO(response.content))

                with self.lock:     # to make the database thread safe
                    image.save(self.path_cache / f"{tile.name}.png")    # TODO can images corrupt when window is closed?
                    expire_time = 60 * 60 * 24 * 7  # 7 days
                    self.cache_json[tile.name]["time"] = time.time() + expire_time
                    self.cache_json[tile.name]["state"] = "loaded"
                    self.viewer.update()

                self.queue.task_done()
            except Exception as e:
                # an error needn't been handled any further because the loading will be retried automatically
                print("Error:", e)

    def get_tile(self, tile):
        """ Request a tile to be loaded.

        Args:
            tile (Tile): tile object of the tile which should be loaded.

        Returns:
            str: path where the slippy tile will be saved after loading from a worker. Returns None if the tile cannot
            exist.
        """
        if not tile.check_existance():
            return None

        with self.lock:
            if tile.name not in self.cache_json:    # load tile if it is not already in the cache database
                # set set state of the tile to loading and save the current time to allow loading retries after a
                # given time
                self.cache_json[tile.name] = {"state": "loading",
                                              "time": time.time()}
                self.queue.put(tile)
            if tile.name in self.cache_json:
                # if the tile is already in the cache database and the loading is not yet finished but a waiting time
                # is exceeded. The loading will be tried again.
                if self.cache_json[tile.name]["state"] == "loading" and \
                        self.cache_json[tile.name]["time"] + config.retry_time_tile < time.time():
                    self.cache_json["time"] = time.time()
                    self.queue.put(tile)
                # reload after reaching expiring date
                if self.cache_json[tile.name]["state"] == "loaded" and \
                        self.cache_json[tile.name]["time"] < time.time():
                    self.cache_json[tile.name] = {"state": "loading",
                                                  "time": time.time()}
                    self.queue.put(tile)

        return str(self.path_cache / f"{tile.name}.png")

    def load_cache_json(self):
        """ Load the cache database which is a json file. If the database does not exists create it with its folder.

        Returns:
            (dict): the cache database as a dict
        """
        path = self.path_cache / "database.json"
        if not path.is_file():
            path.parent.mkdir(parents=True, exist_ok=True)
            with open(path, "w") as json_file:
                json.dump(dict(), json_file)
        with open(path, "r") as json_file:
            return json.load(json_file)

    def save_cache_json(self):
        """ Save cache database back to a file.
        """
        path = self.path_cache / "database.json"
        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "w") as json_file:
            json.dump(self.cache_json, json_file)

    def close(self):
        """ This is triggered when the parent of the object is destroyed. So this should fire if the main window is
        closed and the database will be saved to a file when no worker changes the database.
        """
        # TODO is there a better/safer way?
        with self.lock:
            self.save_cache_json()

    def draw(self, viewer, qpainter, alpha):
        """ Function to draw on a View.

        Args:
            viewer (Viewer): object which must is drawn on and which must be updated
            qpainter (QPainter): object which is used to draw
            alpha (float): opacity to draw
        """
        qpainter.setOpacity(alpha)
        main_tile = Tile(viewer.lat, viewer.lon, viewer.zoom)
        offset_x = int((viewer.x - main_tile.center_x) * viewer.scale_x)
        offset_y = int((viewer.y - main_tile.center_y) * viewer.scale_y)
        num_x = int(np.ceil(int(np.ceil(viewer.frameGeometry().width() / config.image_size) + 1) / 2))
        num_y = int(np.ceil(int(np.ceil(viewer.frameGeometry().height() / config.image_size) + 1) / 2))

        for a in range(-num_x, num_x + 1):
            for b in range(-num_y, num_y + 1):
                tile = Tile.from_num(main_tile.xtile + a, main_tile.ytile + b, main_tile.zoom)
                if not tile.check_existance():
                    continue
                path_image = self.get_tile(tile)

                # try:
                if pathlib.Path(path_image).is_file():
                    pic = QPixmap(str(path_image))
                else:
                    pic = QPixmap(viewer.asset_error_image)

                pic = pic.scaled(config.image_size, config.image_size)
                qpainter.drawTiledPixmap(
                    -offset_x + a * config.image_size + viewer.frameGeometry().width() * 0.5 - config.image_size * 0.5,
                    offset_y + b * config.image_size + viewer.frameGeometry().height() * 0.5 - config.image_size * 0.5,
                    config.image_size, config.image_size, pic)



# -*- coding: utf-8 -*-

import numpy as np

from Osmapy.utils import calc
from Osmapy.utils.config import config


class Tile:
    """ Class to represent a slippy tile.
    """

    def __init__(self, lat, lon, zoom):
        """ Constructer with the latitude and longitude which should lay inside of the slippy tile.

        Args:
            lat (float): latitude of a position which should lay inside of the tile
            lon (float): longitude of a position which should lay inside of the tile
            zoom (int): zoom level of the tile
        """
        self.lat = lat
        self.lon = lon
        self.zoom = zoom

        self.x, self.y = calc.deg2xy(self.lat, self.lon)

        self.xtile, self.ytile = calc.deg2num(self.lat, self.lon, self.zoom)

        self.int_xtile, self.int_ytile = int(self.xtile), int(self.ytile)

        self.bbox_deg = calc.get_bbox_deg(self.xtile, self.ytile, self.zoom)
        self.bbox_xy = calc.get_bbox_xy(self.xtile, self.ytile, self.zoom)

        self.width_x = np.abs(self.bbox_xy.left - self.bbox_xy.right)
        self.width_y = np.abs(self.bbox_xy.top - self.bbox_xy.bottom)

        self.center_x, self.center_y = calc.deg2xy(*calc.num2deg(self.int_xtile + 0.5, self.int_ytile + 0.5, self.zoom))

        self.scale_x = np.abs(self.bbox_xy.left - self.bbox_xy.right) / config.image_size
        self.scale_y = np.abs(self.bbox_xy.top - self.bbox_xy.left) / config.image_size

        self.name = f"{self.int_xtile}_{self.int_ytile}_{self.zoom}"  # internal name of this tile

    @classmethod
    def from_num(cls, xtile, ytile, zoom):
        """ Alternative constructor using the slippy tile numbers.

        Args:
            xtile (float): x slippy tile number of the requested tile
            ytile (float): y slippy tile number of the requested tile
            zoom (int): zoom level of the tile
        """
        lat, lon = calc.num2deg(xtile, ytile, zoom)
        return cls(lat, lon, zoom)

    def check_existance(self):
        """ Checks if the tilenumbers are valid and if the tile can exists on a slippy tile server.

        Returns:
            bool: True if tile can exists, False if not
        """
        # these bounds are given by the mercator projection
        if 0 <= self.xtile <= 2**self.zoom and 0 <= self.ytile <= 2**self.zoom:
            return True
        return False



# -*- coding: utf-8 -*-

from functools import partial

from PySide2.QtWidgets import QFormLayout, QWidget, QLineEdit, QLabel, QPushButton, QInputDialog


class ElementViewer(QWidget):
    """ Widget which contains a TextEdit with the information of a selected OSM element.
    """

    def __init__(self, parent):
        super(ElementViewer, self).__init__()
        self.parent = parent
        self.node = None

        self.clear()

    def clear(self):
        """ Clear ElementViewer to show no node.
        """
        QWidget().setLayout(self.layout())
        layout = QFormLayout()
        text = QLabel("Right Click to Select Node")
        text.setStyleSheet("font-weight: bold")
        layout.addRow(text)
        self.setLayout(layout)

    def set_node(self, node):
        """ Change the node shown in the ElementViewer

        Args:
            node (Node): Node that should be shown in the ElementViewer.
        """
        self.id = node.id

        QWidget().setLayout(self.layout())
        layout = QFormLayout()

        # Node information which cannot be changed by the user
        layout.addRow("Id", QLabel(str(node.id)))
        layout.addRow("Uid", QLabel(str(node.data["uid"])))
        layout.addRow("User", QLabel(str(node.data["user"])))
        layout.addRow("Version", QLabel(str(node.data["version"])))
        layout.addRow("Changeset", QLabel(str(node.data["changeset"])))
        layout.addRow("Timestamp", QLabel(str(node.data["timestamp"])))

        # Properties of nodes which can be changed by the user
        for filed_humanreadable, field in zip(["Latitiude", "Longitude"], ["lat", "lon"]):
            edit = QLineEdit(str(node.data[field]))
            layout.addRow(filed_humanreadable, edit)
            edit.textChanged.connect(partial(self.modify_property, field))

        # Tags which can be deleted an modified
        if len(node.data["tags"]) > 0:
            text = QLabel("Tags")
            text.setStyleSheet("font-weight: bold")
            layout.addRow(text)

            for key, value in node.data["tags"].items():
                key_edit = QPushButton(key)
                value_edit = QLineEdit(value)
                layout.addRow(key_edit, value_edit)
                key_edit.clicked.connect(partial(self.remove_tag, key))
                value_edit.textChanged.connect(partial(self.modify_tag, key))

        button = QPushButton("Add Tag")
        layout.addRow(button)
        button.clicked.connect(self.new_tag)

        button = QPushButton("Delete Node")
        layout.addRow(button)
        button.clicked.connect(self.delete_node)

        self.setLayout(layout)

    def delete_node(self):
        """ Delete the currently selected node.
        """
        del self.parent.elements_loader.elements[self.id]
        self.parent.elements_loader.selected_node = None
        self.parent.viewer.update()
        self.clear()

    def modify_property(self, field, value):
        """ Callback to change a nodes property by changing a text field.

        Args:
            field (str): property which should be changed
            value (str): new value
        """
        # TODO typechecking
        self.parent.elements_loader.elements[self.id].data[field] = value
        self.parent.viewer.update()

    def modify_tag(self, key, value):
        """ Callback to modify a tag

        Args:
            key (str): tag key
            value (str): tag value
        """
        self.parent.elements_loader.elements[self.id].data["tags"][key] = value
        self.parent.viewer.update()

    def remove_tag(self, key):
        """ Remove a tag from an object

        Args:
            key (str): key of the tag which should be removed
        """
        del self.parent.elements_loader.elements[self.id].data["tags"][key]
        self.set_node(self.parent.elements_loader.elements[self.id])

    def new_tag(self):
        """ Ask user for key and value of the new tag and create it.
        """
        key, ok = QInputDialog().getText(self, "New Tag", "Key", QLineEdit.Normal)
        if ok and key:
            value, ok = QInputDialog().getText(self, "New Tag", "Value", QLineEdit.Normal)
            if ok and value:
                self.parent.elements_loader.elements[self.id].data["tags"][key] = value
                self.set_node(self.parent.elements_loader.elements[self.id])



# -*- coding: utf-8 -*-

import lxml.etree as ET
from PySide2.QtWidgets import QDialog, QGridLayout, QLabel, QTextEdit, QLineEdit, QPushButton, QMessageBox

from Osmapy.utils.config import config


class ChangesetForm(QDialog):
    def __init__(self, parent):
        super(ChangesetForm, self).__init__()
        self.parent = parent

        self.status_codes = {200: "Success",
                             400: "Bad request",
                             401: "Login was unsuccessful",
                             404: "Not found",
                             403: "User has been blocked",
                             405: "Method Not Allowed",
                             409: "Conflict"}

    def show(self):
        # reset layout
        QDialog().setLayout(self.layout())

        osm_change = ET.tostring(self.parent.changeset.create_osmChange(-1), pretty_print=True).decode()

        self.setWindowTitle("Upload changes")

        label1 = QLabel("osmChange file:")
        text = QTextEdit()
        text.setReadOnly(True)
        text.setText(osm_change)

        label2 = QLabel("Comment:")
        self.comment = QLineEdit()

        label3 = QLabel("Username:")
        self.username = QLineEdit()

        label4 = QLabel("Password:")
        self.password = QLineEdit()
        self.password.setEchoMode(QLineEdit.Password)

        button = QPushButton("Submit Changes")

        layout = QGridLayout()
        layout.addWidget(label1, 0, 0, 1, 2)
        layout.addWidget(text, 1, 0, 1, 2)
        layout.addWidget(label2, 2, 0, 1, 1)
        layout.addWidget(self.comment, 2, 1, 1, 1)
        layout.addWidget(label3, 3, 0, 1, 1)
        layout.addWidget(self.username, 3, 1, 1, 1)
        layout.addWidget(label4, 4, 0, 1, 1)
        layout.addWidget(self.password, 4, 1, 1, 1)
        layout.addWidget(button, 5, 0, 1, 2)
        self.setLayout(layout)

        if "login_name" in config:
            self.username.setText(config.login_name)
        if "password" in config:
            self.password.setText(config.password)

        button.clicked.connect(self.click)
        super().show()

    def click(self):
        if self.username.text() == "":
            box = QMessageBox()
            box.setText("Please enter your username!")
            box.setIcon(QMessageBox.Icon.Warning)
            box.exec()
            return
        if self.password.text() == "":
            box = QMessageBox()
            box.setText("Please enter your password!")
            box.setIcon(QMessageBox.Icon.Warning)
            box.exec()
            return
        if self.comment.text() == "":
            box = QMessageBox()
            box.setText("Please enter a comment!")
            box.setIcon(QMessageBox.Icon.Warning)
            box.exec()
            return

        status_codes = self.parent.changeset.submit(self.comment.text(), self.username.text(), self.password.text())

        self.parent.viewer.load_elements()
        self.hide()

        if status_codes == 200:
            box = QMessageBox()
            box.setWindowTitle("Success")
            box.setText("Successfully uploaded changes!")
            box.setIcon(QMessageBox.Icon.Information)
            box.exec()
        else:
            box = QMessageBox()
            box.setWindowTitle("ERROR")
            box.setText(f"Error:\n{self.status_codes[status_codes]}")
            box.setIcon(QMessageBox.Icon.Critical)
            box.exec()





import gpxpy
from PySide2 import QtCore
from PySide2.QtCore import QLine
from PySide2.QtGui import QColor, QPen

from Osmapy.utils.Point import Point


class GPXLoader:
    def __init__(self, path):
        """ Class to manage GPX information.

        Args:
            path (pathlib.Path): path to the gpx file

        """
        self.points = []
        try:
            with open(path, "r") as gpx_file:
                gpx = gpxpy.parse(gpx_file)
                self.points = []
                for track in gpx.tracks:
                    for segment in track.segments:
                        for point in segment.points:
                            point = Point(point.latitude, point.longitude)
                            self.points.append(point)
        except Exception as e:
            # TODO error handling
            print(e)

    def draw(self, viewer, qpainter, alpha):
        """ Function to draw on a View.

        Args:
            viewer (Viewer): object which must is drawn on and which must be updated
            qpainter (QPainter): object which is used to draw
            alpha (float): opacity to draw
        """
        qpainter.setOpacity(alpha)
        qpainter.setBrush(QColor(QtCore.Qt.red))
        qpainter.setPen(QPen(QColor(QtCore.Qt.red), 2, QtCore.Qt.SolidLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin))

        points = [viewer.xy2screen(point.x, point.y) for point in self.points]
        pointsA = points[:-1]
        pointsB = points[1:]

        lines = [QLine(pointA[0], pointA[1], pointB[0], pointB[1]) for pointA, pointB in zip(pointsA, pointsB)]
        qpainter.drawLines(lines)



# -*- coding: utf-8 -*-

import numpy as np

from Osmapy.utils import calc


class Point:
    """ Class to manage a point on the earth. Uses latitude, longitude and the meracator coordinates.
    """

    def __init__(self, lat, lon):
        self.lat = lat
        self.lon = lon
        self.x, self.y = calc.deg2xy(lat, lon)

    @classmethod
    def from_deg(cls, lat, lon):
        return cls(lat, lon)

    @classmethod
    def from_xy(cls, x, y):
        lat, lon = calc.xy2deg(x, y)
        return cls(lat, lon)

    def __repr__(self):
        return f"Lat: {self.lat} Lon: {self.lon}"

    def __sub__(self, other):
        x = self.x - other.x
        y = self.y - other.y
        return Point.from_xy(x, y)

    def __add__(self, other):
        x = self.x + other.x
        y = self.y + other.y
        return Point.from_xy(x, y)

    def __abs__(self):
        return np.sqrt(self.x ** 2 + self.y ** 2)



# -*- coding: utf-8 -*-

import lxml.etree as ET
import requests

from Osmapy.utils.config import config


class Changeset:
    """ Represent a changeset. Create it and send it to the server.
    """

    def __init__(self, parent):
        self.parent = parent
        self.osm_api_url = config.osm_api_url

        self.headers = {"User-Agent": config.user_agent, "Content-Type": "application/xml"}

    def submit(self, comment, username, password):
        """ Submit the changeset to the server.

        Args:
            comment (str): comment of this changeset
            username (str): users username
            password (str): users password
        """
        self.username = username
        self.password = password

        status_code = self.create_changeset(comment)
        if status_code != 200:
            return status_code

        osmChange = ET.tostring(self.create_osmChange(self.changeset_id))
        status_code = self.upload_diff(osmChange)
        if status_code != 200:
            return status_code

        status_code = self.close()
        if status_code != 200:
            return status_code

        self.parent.element_viewer.clear()
        self.parent.elements_loader.clear()
        self.parent.viewer.update()

        return status_code

    def create_changeset(self, comment):
        """ Ask server to create a changeset.

        Args:
            comment (str): comment for this changeset

        Returns:
            http status code of the response
        """
        root = ET.Element("osm")
        changeset = ET.SubElement(root, "changeset")
        ET.SubElement(changeset, "tag", k="comment", v=comment)
        tree = ET.ElementTree(root)

        request = f"{config.osm_api_url}/api/0.6/changeset/create"
        result = requests.put(request, data=ET.tostring(tree), headers=self.headers,
                              auth=(self.username, self.password))
        if result.status_code == 200:
            self.changeset_id = int(result.text)
        return result.status_code

    def create_osmChange(self, changeset_id):
        """ Create a osmChange XML file which describes the changes which where perfromed by the user.

        Args:
            changeset_id (int): number of the changeset declared by the server

        Returns:
            xml tree: containing all changes as an xml
        """
        elements_copy = self.parent.elements_loader.elements_copy
        elements = self.parent.elements_loader.elements

        root = ET.Element("osmChange")

        # created
        create = ET.SubElement(root, "create")

        created_node_keys = set(elements) - set(elements_copy)
        created_nodes = [elements[k] for k in created_node_keys]

        for i, node in enumerate(created_nodes):
            create.insert(-1, node.create_xml(-i - 1, changeset_id))

        # deleted
        delete = ET.SubElement(root, "delete")

        deleted_node_keys = set(elements_copy) - set(elements)
        deleted_nodes = [elements_copy[k] for k in deleted_node_keys]

        for i, node in enumerate(deleted_nodes):
            delete.insert(-1, node.create_xml(node.id, changeset_id, tags=False))

        # modified
        modify = ET.SubElement(root, "modify")

        keys = set(elements).intersection(set(elements_copy))
        modified_nodes = [elements[k] for k in keys if elements[k] != elements_copy[k]]

        for i, node in enumerate(modified_nodes):
            modify.insert(-1, node.create_xml(node.id, changeset_id))

        tree = ET.ElementTree(root)

        return tree

    def upload_diff(self, osm_change):
        """ Upload osmChange file to the server

        Args:
            osm_change (str): xml as a string

        Returns:
            http status code of the response
        """
        request = f"{config.osm_api_url}/api/0.6/changeset/{self.changeset_id}/upload"
        result = requests.post(request, data=osm_change, headers=self.headers, auth=(self.username, self.password))
        return result.status_code

    def close(self):
        """ Close the changeset to finish the editing

        Returns:
            http status code of the response
        """
        request = f"{config.osm_api_url}/api/0.6/changeset/{self.changeset_id}/close"
        result = requests.put(request, headers=self.headers, auth=(self.username, self.password))
        return result.status_code



# -*- coding: utf-8 -*-

import pathlib
import webbrowser

import numpy as np
from PySide2 import QtCore
from PySide2.QtGui import QPainter, QColor, QPen
from PySide2.QtWidgets import (QDialog, QApplication)

from Osmapy.GPXLoader.GPXLoader import GPXLoader
from Osmapy.TileLoader import TileLoader, Tile
from Osmapy.Viewer.OSMCopyright import OSMCopyright
from Osmapy.utils import calc
from Osmapy.utils.config import config


class Viewer(QDialog):
    """ Viewer widget where the map is shown with the slippy tiles in the background and OSM objects.
    """

    def __init__(self, parent=None):
        super(Viewer, self).__init__()

        self.tile_loaders = []
        for config_id in range(len(config.slippy_tiles)):
            self.tile_loaders.append(TileLoader.TileLoader(self, config_id))
        self.parent = parent
        self.element_viewer = self.parent.element_viewer

        self.lat = config.start_latitude
        self.lon = config.start_longitude
        self.zoom = config.start_zoom
        self.x, self.y = calc.deg2xy(self.lat, self.lon)

        # scale
        tile = Tile.Tile(self.lat, self.lon, self.zoom)
        self.scale_x = config.image_size / tile.width_x
        self.scale_y = config.image_size / tile.width_y

        self.click = False

        self.elements_loader = self.parent.elements_loader

        for tile_loader in self.tile_loaders:
            self.destroyed.connect(tile_loader.close)

        path_base = pathlib.Path(__file__).parent
        self.asset_error_image = str(path_base / pathlib.Path("../../assets/error.png"))

        self.osm_copyright = OSMCopyright()

        self.setAcceptDrops(True)  # allow file dropping

        self.layers = self.parent.layer_manager
        for config_id, tile_loader in enumerate(self.tile_loaders):
            self.layers.add_layer(tile_loader, config.slippy_tiles[config_id].name)
        self.layers.add_layer(self.elements_loader, "OSM Nodes")

        self.mode = "normal"  # mode for clicking events

    def set_deg(self, lat, lon):
        """ Set center of the view.

        Args:
            lat (float): latitude of the center
            lon (float): longitude of the center
        """
        self.lat = lat
        self.lon = lon
        self.x, self.y = calc.deg2xy(self.lat, self.lon)
        self.set_zoom(self.zoom)

    def set_xy(self, x, y):
        """ Set center of the view.

        Args:
            x (float): mercator x of the center
            y (float): mercator y of the center
        """
        self.x = x
        self.y = y
        self.lat, self.lon = calc.xy2deg(self.x, self.y)
        self.set_zoom(self.zoom)

    def set_zoom(self, zoom):
        """ Set zoom level of the view.

        Args:
            zoom (int): zoom level
        """
        self.zoom = zoom
        self.parent.statusBar().showMessage(f"Lat: {self.lat:.7f} Lon: {self.lon:.7f} Zoom: {zoom}")
        tile = Tile.Tile(self.lat, self.lon, self.zoom)
        self.scale_x = config.image_size / tile.width_x
        self.scale_y = config.image_size / tile.width_y

    def screen2xy(self, xscreen, yscreen):
        """ Convert from screen coordinates to mercator coordinates.

        Args:
            xscreen (float): x coordinate on view
            yscreen (float): y coordinate on view

        Returns:
            (float, float): mercator x and y
        """
        xscreen -= self.frameGeometry().width() / 2
        yscreen -= self.frameGeometry().height() / 2
        yscreen *= -1
        xscreen /= self.scale_x
        yscreen /= self.scale_y

        xscreen += self.x
        yscreen += self.y

        return xscreen, yscreen

    def xy2screen(self, x, y):
        """ Convert mercator x and y to screen coordinates.

        Args:
            x (float): mercator x
            y (float): mercator y

        Returns:
            (float, float): coordinates on the view
        """
        xscreen = (x - self.x) * self.scale_x + self.frameGeometry().width() / 2
        yscreen = -(y - self.y) * self.scale_y + self.frameGeometry().height() / 2
        return xscreen, yscreen

    def paintEvent(self, event):
        """ This is triggered on every update. All the connected layers are drawn here. Also the copyright information
        is added here.

        Args:
            event (Event): not yet used
        """
        qpainter = QPainter(self)
        qpainter.setRenderHint(QPainter.Antialiasing)

        for layer, alpha in self.layers.get_layers():
            layer.draw(self, qpainter, alpha)

        qpainter.setBrush(QColor(0, 0, 0, 0))
        qpainter.setPen(QPen(QColor(QtCore.Qt.black), 1))
        size = 4
        qpainter.drawRect(- size / 2 + self.frameGeometry().width() * 0.5,
                          - size / 2 + self.frameGeometry().height() * 0.5, size, size)

        # draw OSM information
        self.osm_copyright.draw(self, qpainter)

    def wheelEvent(self, event):
        """ Callback when the mouse wheel is used. Here the zooming is realized.

        Args:
            event (Event): scrolling event includes the amount of scrolling
        """
        if abs(event.delta()) != 0:
            delta = event.delta() // abs(event.delta())
            if delta == 1 and self.zoom < 19:
                self.set_zoom(self.zoom + 1)

                self.update()
            if delta == -1 and self.zoom > 0:
                self.set_zoom(self.zoom - 1)
                self.update()

    def mouseMoveEvent(self, event):
        """ Callback when the mouse is moved. Here the dragging of the map is realized.

        Args:
            event (Event): contains the position of the mouse
        """
        if event.buttons() == QtCore.Qt.LeftButton:
            if not self.click:
                self.start = event.globalPos()
                self.click = True
                self.start_x = self.x
                self.start_y = self.y
            else:
                moved = event.globalPos() - self.start
                self.set_xy(self.start_x - moved.x() / self.scale_x, self.start_y + moved.y() / self.scale_y)
                self.update()

    def mouseReleaseEvent(self, event):
        """ Callback when the mouse is released. This is needed to realize the dragging.

        Args:
            event (Event): not yet used
        """
        self.click = False

    def mousePressEvent(self, event):
        """ Callback when the mouse is clicked. This is use to react on a click on the OSM copyright, select a node
        or create a new node.

        Args:
            event (Event): contains the click information
        """
        if event.buttons() == QtCore.Qt.LeftButton:
            # click on copyright
            if event.x() > self.frameGeometry().width() - self.osm_copyright.width - self.osm_copyright.margin and \
                    event.y() > self.frameGeometry().height() - self.osm_copyright.height - self.osm_copyright.margin:
                webbrowser.open(self.osm_copyright.url)
                return

        if self.mode == "normal":
            if event.buttons() == QtCore.Qt.RightButton:
                posx = event.x()
                posy = event.y()
                smallest = 9999999
                elem_id = None
                for key, elem in self.elements_loader.elements.items():
                    xscreen, yscreen = self.xy2screen(elem.x, elem.y)
                    dist = np.sqrt((xscreen - posx) ** 2 + (yscreen - posy) ** 2)
                    if dist < smallest:
                        elem_id = key
                        smallest = dist
                if elem_id:
                    self.elements_loader.selected_node = elem_id
                    self.update()
                    self.element_viewer.set_node(self.elements_loader.elements[elem_id])
        elif self.mode == "new_node":
            if event.buttons() == QtCore.Qt.RightButton:
                x, y = self.screen2xy(event.x(), event.y())
                lat, lon = calc.xy2deg(x, y)
                self.elements_loader.new_node(lat, lon)
                self.update()
                self.change_mode("normal")

    def dragEnterEvent(self, event):
        """ This callback is fired when something is dragged above the view. It is shown to the user that this is
        accepted to allow dropping GPX files.

        Args:
            event (Event): information of the dragged object
        """

        if event.mimeData().hasUrls():
            event.acceptProposedAction()

    def dropEvent(self, event):
        """ Event when something is dropped into the view. This is used to read GPX files.

        Args:
            event (Event): includes the dropping metadata
        """
        path = pathlib.Path(event.mimeData().urls()[0].toLocalFile())
        self.layers.add_layer(GPXLoader(path), event.mimeData().urls()[0].toLocalFile())

        self.update()

    def keyPressEvent(self, event):
        """ Callback for keypress events. This is used to move the selected object around. this scales with the zoom
        level.

        Args:
            event (Event): contains the information of the pressed key
        """
        if self.parent.elements_loader.selected_node:
            node_id = self.parent.elements_loader.selected_node
            node = self.parent.elements_loader.elements[node_id]
            if event.key() == QtCore.Qt.Key_Right:
                node.set_position(node.x + 1 / self.scale_x, node.y)
                self.element_viewer.set_node(self.elements_loader.elements[node_id])
                self.update()
            if event.key() == QtCore.Qt.Key_Left:
                node.set_position(node.x - 1 / self.scale_x, node.y)
                self.element_viewer.set_node(self.elements_loader.elements[node_id])
                self.update()
            if event.key() == QtCore.Qt.Key_Up:
                node.set_position(node.x, node.y + 1 / self.scale_y)
                self.element_viewer.set_node(self.elements_loader.elements[node_id])
                self.update()
            if event.key() == QtCore.Qt.Key_Down:
                node.set_position(node.x, node.y - 1 / self.scale_y)
                self.element_viewer.set_node(self.elements_loader.elements[node_id])
                self.update()

    def load_elements(self):
        """ Start loading OSM elements from the api which belong in the current map view.
        """
        left, top = self.screen2xy(0, 0)
        right, bottom = self.screen2xy(self.frameGeometry().width(), self.frameGeometry().height())
        north, west = calc.xy2deg(left, bottom)
        south, east = calc.xy2deg(right, top)
        self.elements_loader.load(west, north, east, south)

        self.update()

    def undo_changes(self):
        """ Undo the changes of the nodes.
        """
        self.parent.element_viewer.clear()
        self.parent.elements_loader.clear()
        self.update()
        self.load_elements()

    def change_mode(self, mode):
        """ Changes what happens when the mouse is clicked in the view. Also changes the cursor style.

        Args:
            mode (str): humanreadable mode code
        """
        self.parent.setToolTip("")
        if mode == "new_node":
            self.parent.setToolTip("Right Click to create Node")
            QApplication.setOverrideCursor(QtCore.Qt.CrossCursor)
        if mode == "normal":
            QApplication.restoreOverrideCursor()
        self.mode = mode



# -*- coding: utf-8 -*-
from PySide2 import QtCore
from PySide2.QtWidgets import QWidget, QGridLayout, QListWidget, QAbstractItemView, QSlider, QLabel


class Layers(QListWidget):
    """ Class for layer widget to allow update of viewer when order of layers is changed.
    """
    def __init__(self, viewer):
        super(Layers, self).__init__()
        self.viewer = viewer

        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)

    def dropEvent(self, event):
        """ Override to update viewer when order changes.
        """
        super(QListWidget, self).dropEvent(event)
        self.viewer.update()



class LayerManager(QWidget):
    """ Class to manage the layers of a viewer. Allow to change the order of layers by drag and drop.
    """
    def __init__(self, viewer):
        super(LayerManager, self).__init__()


        self.selected_layer = None

        self.viewer = viewer
        self.layers = dict()

        self.layer_widget = Layers(self.viewer)
        self.alpha_slider = QSlider(QtCore.Qt.Horizontal)
        self.alpha_label = QLabel(f"Select Layer to change Opacity")

        layout = QGridLayout()
        layout.addWidget(self.layer_widget, 0, 0, 1, 1)
        layout.addWidget(self.alpha_label, 1, 0, 1, 1)
        layout.addWidget(self.alpha_slider, 2, 0, 1, 1)
        self.setLayout(layout)

        self.layer_widget.itemClicked.connect(self.select_layer)
        self.alpha_slider.valueChanged.connect(self.slider_changed)
        self.alpha_slider.setValue(99)



    def add_layer(self, layer, name):
        """ Add new layer to be managed.

        Args:
            layer (Object): object that implements a draw function.
            name (str): String which is the name of the layer.
        """
        self.layers[name] = {"layer": layer,
                             "alpha": 1}

        self.layer_widget.addItem(name)

    def get_layers(self):
        """ Get list of layers representing the order of the LayerManager in the UI.

        Returns:
            [Objects]: objects which implement a draw function
        """
        names = [self.layer_widget.item(i).data(0) for i in range(self.layer_widget.count())]
        result = [(self.layers[name]["layer"], self.layers[name]["alpha"]) for name in names]
        return result

    def select_layer(self, item):
        """ Select a layer to change opacity.

        Args:
            item: item of the list
        """
        name = item.data(0)
        self.selected_layer = name
        self.alpha_label.setText(f"Opacity ({self.selected_layer}):")

        # set slider to current value
        value = self.layers[name]["alpha"]
        value = value*100 - 1
        self.alpha_slider.setValue(value)

    def slider_changed(self, value):
        """ Callback when the opacity slider was moved.

        Args:
            value (int): slider value 0-99
        """
        value = (value+1) / 100
        if self.selected_layer in self.layers:
            self.layers[self.selected_layer]["alpha"] = value
        self.viewer.update()




# -*- coding: utf-8 -*-
import os
import sys
from functools import partial
from subprocess import call

from PySide2 import QtCore
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import (QApplication, QMainWindow, QToolBar, QDockWidget)

from Osmapy.Changeset.Changeset import Changeset
from Osmapy.Changeset.ChangesetForm import ChangesetForm
from Osmapy.ElementsLoader.ElementsLoader import ElementsLoader
from Osmapy.Viewer import Viewer
from Osmapy.Viewer.ElementViewer import ElementViewer
from Osmapy.Viewer.LayerManager import LayerManager
from Osmapy.utils import config


class Main(QMainWindow):
    """ MainWindow which contains all widgets of Osmapy.
    """

    def __init__(self, parent=None):
        super(Main, self).__init__(parent)
        self.setWindowTitle("Osmapy")
        # All widgets should be destryed when the main window is closed. This the widgets can use the destroyed widget
        # to allow clean up. E.g. save the database of the TileLoader.
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        self.resize(config.config.window_size[0], config.config.window_size[1])

        self.elements_loader = ElementsLoader()

        # Element Viewer as DockWidget
        self.element_viewer = ElementViewer(self)
        self.dock_element_viewer = QDockWidget()
        self.dock_element_viewer.setWindowTitle("Element Viewer")
        self.dock_element_viewer.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        self.dock_element_viewer.setWidget(self.element_viewer)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.dock_element_viewer)

        # LayerManger as DockWidget
        self.layer_manager = LayerManager(self)
        self.dock_layer_manager = QDockWidget()
        self.dock_layer_manager.setWindowTitle("Layer Manager")
        self.dock_layer_manager.setFeatures(QDockWidget.DockWidgetFloatable | QDockWidget.DockWidgetMovable)
        self.dock_layer_manager.setWidget(self.layer_manager)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.dock_layer_manager)

        self.viewer = Viewer.Viewer(self)
        self.setCentralWidget(self.viewer)
        self.viewer.setFocus()
        self.viewer.setFocusPolicy(QtCore.Qt.StrongFocus)

        self.changeset = Changeset(self)
        self.changset_form = ChangesetForm(self)

        self.toolbar = QToolBar()
        self.toolbar.addAction("Load Elements", self.viewer.load_elements)
        self.toolbar.addAction("Undo Changes", self.viewer.undo_changes)
        self.toolbar.addAction("Create Node", partial(self.viewer.change_mode, "new_node"))
        self.toolbar.addAction("Upload Changes", self.changset_form.show)
        if os.name == "nt":
            self.toolbar.addAction("Open Configuration", partial(os.startfile, str(config.path_config)))
        elif sys.platform == "darwin":
            self.toolbar.addAction("Open Configuration", partial(call, ["open", str(config.path_config)]))
        else:
            self.toolbar.addAction("Open Configuration", partial(call, ["xdg-open", str(config.path_config)]))
        self.addToolBar(self.toolbar)

        self.statusBar().showMessage("Welcome to Osmapy!")


def main():
    # Staring point of Osmapy
    app = QApplication()
    app.setWindowIcon(QIcon("./assets/appicon.png"))
    app.setApplicationName("Osmapy")
    main_window = Main()
    main_window.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

""" In this file all the necessary calculations should be defined.
"""

from dataclasses import dataclass

import numpy as np


@dataclass
class Bbox_deg:
    """ Class for a bounding box.
    """
    left: float
    top: float
    right: float
    bottom: float


@dataclass
class Bbox_xy:
    """ Class for a bounding box.
    """
    left: float
    top: float
    right: float
    bottom: float


def xy2deg(x, y):
    """ Calculate the latitude and longitude in degree given the mercator x and y.

    Args:
        x (float): mercator x
        y (float): mercator y

    Returns:
        (float, float): tuple with the latitude and longitude in degree
    """
    return (180.0 / np.pi * (2.0 * np.arctan(np.exp(y * np.pi / 180.0)) - np.pi / 2.0), x)


def deg2xy(lat, lon):
    """ Calculate the mercator x and y given the longitude and latitiude in degree.

    Args:
        lat (float): latitude in degree
        lon (float): longitude in degree

    Returns:
        (float, float): tuple with the mercator x and y
    """
    return (lon, 180.0 / np.pi * np.log(np.tan(np.pi / 4.0 + lat * (np.pi / 180.0) / 2.0)))


def deg2num(lat, lon, zoom):
    """ Calculate the slippy tile numbers given a latitude and longitude in degree and the zoom level.

    Args:
        lat (float): latitude in degree
        lon (float): longitude in degree
        zoom (int): zoom level of the tile

    Returns:
        (float, float): tuple with the x and y of the tile. Must be converted to integer to use it with a tile
                        server
    """
    lat_rad = np.radians(lat)
    n = 2.0 ** zoom
    xtile = (lon + 180.0) / 360.0 * n
    ytile = (1.0 - np.arcsinh(np.tan(lat_rad)) / np.pi) / 2.0 * n
    return (xtile, ytile)


def num2deg(xtile, ytile, zoom):
    """ Calculate the latitude and longitude in degree given the slippy tile numbers.

    Args:
        xtile (float): x of the tile
        ytile (float): y of the tile
        zoom (int): zoom level of the tile

    Returns:
        (float, float): tuple with the latitude and longitude in degree
    """
    n = 2.0 ** zoom
    lon = xtile / n * 360.0 - 180.0
    lat_rad = np.arctan(np.sinh(np.pi * (1 - 2 * ytile / n)))
    lat = np.degrees(lat_rad)
    return (lat, lon)


def get_bbox_deg(xtile, ytile, zoom):
    """ Calculate bounding box of a tile.

    Args:
        xtile (float): x of the tile
        ytile (float): y of the tile
        zoom (int): zoom level of the tile

    Returns:
        (Bbox): dataclass with the bounding box
    """
    top, left = num2deg(xtile, ytile, zoom)
    bottom, right = num2deg(xtile + 1, ytile + 1, zoom)
    bbox = Bbox_xy(left=left, top=top, right=right, bottom=bottom)
    return bbox


def get_bbox_xy(xtile, ytile, zoom):
    """ Calculate bounding box of a tile.

    Args:
        xtile (float): x of the tile
        ytile (float): y of the tile
        zoom (int): zoom level of the tile

    Returns:
        (Bbox): dataclass with the bounding box
    """
    top, left = deg2xy(*num2deg(xtile, ytile, zoom))
    bottom, right = deg2xy(*num2deg(xtile + 1, ytile + 1, zoom))
    bbox = Bbox_xy(left=left, top=top, right=right, bottom=bottom)
    return bbox



# -*- coding: utf-8 -*-

from string import Template

import requests
from PySide2 import QtCore
from PySide2.QtGui import QColor, QPen
from PySide2.QtWidgets import QMessageBox

from Osmapy.ElementsLoader import Node
from Osmapy.utils import calc
from Osmapy.utils.config import config


class ElementsLoader:
    """ This class provides a loader for OSM elements from the OSM server.
    """

    def __init__(self):
        self.elements_copy = dict()  # copy of elements to find changes
        self.elements = dict()
        self.headers = {"Accept": "application/json", "User-Agent": config.user_agent}

        self.selected_node = None
        self.new_node_counter = -1

    def clear(self):
        """ Reset the elements dicts and the counter
        """
        self.selected_node = None
        self.new_node_counter = -1
        self.elements_copy = dict()
        self.elements = dict()

    def load(self, west, north, east, south):
        """ This function loads all node elements from a given bounding box. The function returns all nodes loaded with
        this object so far.

        Args:
            west (float): longitude of the bounding box in degree
            north (float): latitude of the bounding box in degree
            east (float): longitude of the bounding box in degree
            south (float): latitude of the bounding box in degree

        Returns:
            {Node}: dict of all OSM nodes. The keys are the IDs of the nodes.
        """
        url = config.osm_api_url + "/api/0.6/map?bbox=${west},${north},${east},${south}"
        request = Template(url)
        request = request.substitute(west=west, north=north, east=east, south=south)
        result = requests.get(request, headers=self.headers)

        if result.ok:
            result_json = result.json()
            nodes_copy = {raw["id"]: Node.Node(raw) for raw in result_json["elements"].copy() if raw["type"] == "node"}
            self.elements_copy = {**self.elements_copy, **nodes_copy}
            nodes = {raw["id"]: Node.Node(raw) for raw in result_json["elements"].copy() if raw["type"] == "node"}
            self.elements = {**self.elements, **nodes}  # merge old and new nodes
        else:
            box = QMessageBox()
            box.setWindowTitle("Error")
            box.setText("Maybe you have to zoom in because there are to many objects in this area")
            box.setIcon(QMessageBox.Icon.Warning)
            box.exec()

    def new_node(self, lat, lon):
        """ Add new node to the elements list.

        Args:
            lat (float): latitude of the new node
            lon (float): longitude of the new node
        """
        self.elements[self.new_node_counter] = Node.Node.create_new_node(self.new_node_counter, lat, lon)
        self.new_node_counter -= 1

    def draw(self, viewer, qpainter, alpha):
        """ Function to draw on a View.

        Args:
            viewer (Viewer): object which must is drawn on and which must be updated
            qpainter (QPainter): object which is used to draw
            alpha (float): opacity to draw
        """
        qpainter.setOpacity(alpha)
        for elem in self.elements.values():
            qpainter.setBrush(QColor(QtCore.Qt.blue))
            qpainter.setPen(QPen(QColor(QtCore.Qt.black), 1))
            if elem.data["type"] == "node":
                x, y = calc.deg2xy(float(elem.data["lat"]), float(elem.data["lon"]))
                xscreen, yscreen = viewer.xy2screen(x, y)

                size = 6
                qpainter.drawEllipse(xscreen - size / 2, yscreen - size / 2, size, size)

                if self.selected_node and elem.id == self.selected_node:
                    qpainter.setBrush(QColor(0, 0, 0, 0))
                    qpainter.setPen(QPen(QColor(QtCore.Qt.red), 2))
                    size = 10
                    qpainter.drawRect(xscreen - size / 2, yscreen - size / 2, size, size)



import setuptools

setuptools.setup(name="Osmapy",
                 author="Philipp Rigoll",
                 version="0.0.1.1",
                 author_email="philipp@rigoll.de",
                 url="https://github.com/PhilippRigoll/Osmapy",
                 packages=setuptools.find_packages(),
                 include_package_data=True,
                 install_requires=[
                     "numpy",
                     "Pillow",
                     "PyYAML",
                     "PySide2",
                     "easydict",
                     "lxml",
                     "requests",
                     "gpxpy",
                     "dataclasses;python_version<'3.7'"
                 ],
                 python_requires=">=3.6",
                 entry_points={
                     "console_scripts": ["Osmapy=Osmapy.main:main"]
                 })



# -*- coding: utf-8 -*-

""" A yaml configuration file is loaded and converted to a dict which allows the use of the dot operator. Some
configurations are hardcoded here.
"""

import pathlib

import yaml
from easydict import EasyDict

path_base = pathlib.Path(__file__).parent
path_config = path_base / pathlib.Path("../config.txt")

with open(path_config, "r") as file:
    config = EasyDict(yaml.safe_load(file))

config.image_size = 256     # tile size
config.retry_time_tile = 4  # Wait 4 seconds before retry to load a slippy tile



# -*- coding: utf-8 -*-
from PySide2.QtCore import QRect
from PySide2.QtGui import QBrush, QColor, QStaticText, QFont


class OSMCopyright:
    """ Class for the OSM copyright in the right bottom corner of the map.
    """

    def __init__(self):
        self.margin = 2
        self.url = "https://www.openstreetmap.org/copyright"

        self.copyright_text = QStaticText("© OpenStreetMap contributors")
        self.font = QFont()
        self.font.setPointSize(8)
        self.copyright_text.prepare(font=self.font)
        size = self.copyright_text.size()
        self.width, self.height = size.width() + self.margin, size.height() + self.margin

    def draw(self, viewer, qpainter):
        """ Function to draw on a View.

        Args:
            viewer (Viewer): object which must is drawn on and which must be updated
            qpainter (QPainter): object which is used to draw
        """
        qpainter.setFont(self.font)
        self.rect = QRect(viewer.frameGeometry().width() - self.width - self.margin,
                          viewer.frameGeometry().height() - self.height - self.margin,
                          self.width + self.margin,
                          self.height + self.margin)
        qpainter.fillRect(self.rect, QBrush(QColor(255, 255, 255)))

        qpainter.drawStaticText(viewer.frameGeometry().width() - self.width,
                                viewer.frameGeometry().height() - self.height,
                                self.copyright_text)



# -*- coding: utf-8 -*-

import lxml.etree as ET

from Osmapy.utils import calc


class Node:
    """ Class to represent an OSM node.
    """

    def __init__(self, raw):
        """ The constructer uses the raw dictionary of the OSM server answer.

        Args:
            raw (dict): OSM server result for this object.
        """
        self.raw = raw.copy()
        self.id = self.raw["id"]

        self.data = dict(id=str(self.raw["id"]),
                         uid=str(self.raw["uid"]),
                         user=str(self.raw["user"]),
                         version=str(self.raw["version"]),
                         changeset=str(self.raw["changeset"]),
                         timestamp=str(self.raw["timestamp"]),
                         type="node",
                         lat=str(self.raw["lat"]),
                         lon=str(self.raw["lon"]))

        if "tags" in self.raw:
            self.data["tags"] = self.raw["tags"].copy()
        else:
            self.data["tags"] = dict()

        self.x, self.y = calc.deg2xy(self.raw["lat"], self.raw["lon"])
        self.trigger = False

    @classmethod
    def create_new_node(cls, id, lat, lon):
        """ Create a new node and add it to the elements list

        Args:
            id (int): id of the new element
            lat (float): latitude of the new node
            lon (float): longitude of the new node

        Returns:
            Node: new node object
        """
        new_raw = dict(id=id,
                       uid="-1",
                       user="-1",
                       version="0",
                       changeset="-1",
                       timestamp="-1",
                       type="node",
                       lat=lat,
                       lon=lon,
                       tags=dict())
        return cls(new_raw)

    def create_xml(self, id, changeset=None, tags=True):
        """ Create XML representation of the node. Can be used to create a osmChange file.

        Args:
            id (int): id which should be shown in the XML
            tags (bool): the tags should be omitted when deleting a node
        """
        if not changeset:
            changeset = self.data["changeset"]

        xml_node = ET.Element("node",
                              id=str(id),
                              changeset=str(changeset),
                              version=str(self.data["version"]),
                              lat=str(self.data["lat"]),
                              lon=str(self.data["lon"]))

        if tags:
            for key, value in self.data["tags"].items():
                ET.SubElement(xml_node, "tag", k=key, v=value)

        return xml_node

    def set_position(self, x, y):
        """ Set new position of node

        Args:
            x (float): mercator x
            y (float): mercator y
        """
        self.x = x
        self.y = y
        lat, lon = calc.xy2deg(x, y)
        self.data["lat"], self.data["lon"] = str(lat), str(lon)

    def __str__(self):
        """ XML representation of the node.

        Returns:
            str: XML representation as a string
        """
        return ET.tostring(self.create_xml(self.id)).decode()

    def __eq__(self, other):
        """ Compare two node objects, by their string XML representation.

        Args:
            other (Node): node to compare

        Returns:
            bool
        """
        return ET.tostring(self.create_xml(1)).decode() == ET.tostring(other.create_xml(1)).decode()

    def __ne__(self, other):
        """ Compare two node objects, by their string XML representation.

        Args:
            other (Node): node to compare

        Returns:
            bool
        """
        return not self.__eq__(other)
